import os
import random
import ctypes
import time

import h5py
import numpy as np
import pyglet
from pyglet.gl import *

from GramophoneTools.LinMaze import Rule
from GramophoneTools.LinMaze.Tools.Stopwatch import Stopwatch
from GramophoneTools.LinMaze.Tools.filehandler import select_file
import GramophoneTools.Comms as Comms

class LinMazeError(Exception):
    """
    Generic Exception for LinMaze related errors.
    """
    pass

class VRWindow(pyglet.window.Window):
    '''
    A pyglet window that can display VR on a given screen.
    
    :param session: The session that is played in this window.
    :type session: Session

    :param screen_number: Which monitor the window should display on.
    :type screen_number: int

    :param mirrored: True if the contents of this window should be 
        mirrored horizontally. False by default.
    :type mirrored: bool

    :param fullscreen: True if the window should be fullscreen. True by default.
    :type fullscreen: bool
    
    '''

    def __init__(self, session, screen_number,
                 mirrored=False, fullscreen=True):

        self.session = session
        self.mirrored = mirrored
        disp = pyglet.window.Display()
        screens = disp.get_screens()
        screen_number -= 1
        super().__init__(self.session.level.screen_width,
                         self.session.level.screen_height,
                         screen=screens[screen_number],
                         resizable=False, vsync=True,
                         fullscreen=fullscreen, visible=False)

        self.set_mouse_visible(False)
        self.set_caption('pyVR - Monitor #' + str(screen_number + 1))

        dir = os.path.dirname(__file__)
        icon = pyglet.image.load(dir+'\\res\\icon.png')
        self.set_icon(icon)

        # OpenGL init
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        glOrtho(0, self.session.level.screen_width, 0,
                self.session.level.screen_height, -1, 1)
        glMatrixMode(GL_MODELVIEW)
        glDisable(GL_DEPTH_TEST)
        glClearColor(0.0, 0.0, 0.0, 0.0)
        glEnable(GL_TEXTURE_2D)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)

    def on_key_press(self, symbol, modifiers):
        if symbol == pyglet.window.key.SPACE:
            if self.session.paused:
                self.session.unpause()
            else:
                self.session.pause()

        if symbol == pyglet.window.key.ESCAPE:
            pyglet.app.exit()

        if symbol == pyglet.window.key.R:
            self.session.teleport(0)

        if symbol == pyglet.window.key.M:
            self.minimize()

        if symbol == pyglet.window.key.R:
            self.session.gramophone.reset()

    def on_close(self):
        pyglet.app.exit()

    def on_draw(self):
        self.switch_to()

        # Clean canvas
        glClear(GL_COLOR_BUFFER_BIT)

        # Dispay all vr units
        for vru in self.session.vr_units:
            sy = self.session.level.screen_height
            sx = vru["length"]

            glLoadIdentity()
            glColor3f(1.0, 1.0, 1.0)
            glBindTexture(GL_TEXTURE_2D, vru["texture_id"])

            if self.mirrored:
                glTranslatef((-1) * vru["position"] +
                             self.session.level.screen_width -
                             vru["length"], 0, 0)

                glBegin(GL_QUADS)

                glTexCoord2f(0, 0)
                glVertex2i(sx, 0)

                glTexCoord2f(1, 0)
                glVertex2i(0, 0)

                glTexCoord2f(1, 1)
                glVertex2i(0, sy)

                glTexCoord2f(0, 1)
                glVertex2i(sx, sy)
            else:
                glTranslatef(vru["position"], 0, 0)

                glBegin(GL_QUADS)

                glTexCoord2f(1, 0)
                glVertex2i(sx, 0)

                glTexCoord2f(0, 0)
                glVertex2i(0, 0)

                glTexCoord2f(0, 1)
                glVertex2i(0, sy)

                glTexCoord2f(1, 1)
                glVertex2i(sx, sy)

            glEnd()

        if self.session.offset_arrow:
            glLoadIdentity()
            glColor3f(1.0, 0.0, 0.0)
            glTranslatef(self.session.level.zone_offset, 0, 0)
            glBindTexture(GL_TEXTURE_2D, 0)
            glBegin(GL_TRIANGLES)
            glVertex2i(-50, 0)
            glVertex2i(50, 0)
            glVertex2i(0, 150)
            glEnd()

        glFlush()


class VRLog(object):
    """
    A logger for LinMaze Sessions.
    
    :param session: The session that should be logged.
    :type session: Session
    """
    def __init__(self, session):
        self.session = session

        # Make headers
        self.zone_types = list(
            set([zone.zone_type for zone in session.level.zones]))

        self.vrl = h5py.File(session.filename, "w")

        self.vrl.attrs['level_name'] = session.level.name
        self.vrl.attrs['start_time'] = session.start_time
        self.vrl.attrs['start_time_hr'] = session.start_time_hr
        self.vrl.attrs['runtime_limit'] = str(session.runtime_limit)
        self.vrl.attrs['screen_width'] = session.level.screen_width
        self.vrl.attrs['screen_height'] = session.level.screen_height
        self.vrl.attrs['zone_offset'] = session.level.zone_offset
        self.vrl.attrs['transition_width'] = session.level.transition_width
        self.vrl.attrs['RGB'] = session.level.rgb
        self.vrl.attrs['left_monitor'] = str(session.left_monitor)
        self.vrl.attrs['right_monitor'] = str(session.right_monitor)
        self.vrl.attrs['gramophone_serial'] = str(session.gramophone_serial)
        self.vrl.attrs['velocity_ratio'] = session.vel_ratio

        self.vrl.create_dataset("time", (0,),
                                maxshape=(None,), dtype=np.float64)
        self.vrl.create_dataset("g_time", (0,),
                                maxshape=(None,), dtype=np.uint64)
        self.vrl.create_dataset("velocity", (0,),
                                maxshape=(None,), dtype=np.int8)
        self.vrl.create_dataset("position", (0,),
                                maxshape=(None,), dtype=np.uint64)
        self.vrl.create_dataset("teleport", (0,),
                                maxshape=(None,), dtype=np.int8)
        self.vrl.create_dataset("paused", (0,),
                                maxshape=(None,), dtype=np.int8)
        self.vrl.create_dataset("input_1", (0,),
                                maxshape=(None,), dtype=np.int8)
        self.vrl.create_dataset("input_2", (0,),
                                maxshape=(None,), dtype=np.int8)
        self.vrl.create_dataset("output_1", (0,),
                                maxshape=(None,), dtype=np.int8)
        self.vrl.create_dataset("output_2", (0,),
                                maxshape=(None,), dtype=np.int8)
        self.vrl.create_dataset("output_3", (0,),
                                maxshape=(None,), dtype=np.int8)
        self.vrl.create_dataset("output_4", (0,),
                                maxshape=(None,), dtype=np.int8)

        zone_count = len(session.level.zones)
        self.vrl.create_dataset("zone", (0, zone_count), maxshape=(
            None, zone_count), dtype=np.int8)

        for zone_type in self.zone_types:
            self.vrl.create_dataset(
                "zone_types/" + zone_type, (0,),
                maxshape=(None,),
                dtype=np.int8)

        # Make log lists
        self.time_record = []
        self.g_time_record = []
        self.vel_record = []
        self.pos_record = []
        self.teleport_record = []
        self.pause_record = []
        self.input_record_1 = []
        self.input_record_2 = []

        self.output_record_1 = []
        self.output_record_2 = []
        self.output_record_3 = []
        self.output_record_4 = []

        self.zone_id_records = np.empty((0, zone_count), dtype=bool)
        self.zone_type_records = {zt: [] for zt in self.zone_types}

    def make_entry(self, vel, g_time, in_1, in_2, out_1, out_2, out_3, out_4):
        '''
        Makes an entry in all the session logs.

        :param vel: The velocity that sould be logged for this entry.
        :type vel: int

        :param g_time: The internal clock value of the Gramophone.
        :type g_time: int

        :param in_1: The state of digital input 1.
        :type in_1: int

        :param in_2: The state of digital input 2.
        :type in_2: int
        '''
        self.time_record.append(self.session.runtime.value())
        self.g_time_record.append(g_time)
        self.vel_record.append(-vel)
        self.pos_record.append(self.session.virtual_position)

        for zone_type in self.zone_types:
            self.zone_type_records[zone_type].append(
                int(self.session.current_zone.zone_type == zone_type))

        zone_row = [int(self.session.current_zone.zone_id == zone.zone_id)
                    for zone in self.session.level.zones]
        self.zone_id_records = np.append(
            self.zone_id_records, [zone_row], axis=0)

        self.teleport_record.append(int(self.session.teleported))
        self.session.teleported = False

        self.input_record_1.append(in_1)
        self.input_record_2.append(in_2)
        self.output_record_1.append(out_1)
        self.output_record_2.append(out_2)
        self.output_record_3.append(out_3)
        self.output_record_4.append(out_4)

        self.pause_record.append(int(self.session.paused))

        if len(self.time_record) >= 600:
            self.flush_all()

    def flush_all(self):
        ''' Writes all temporary data to file. '''

        self.flush(self.time_record, 'time')
        self.flush(self.g_time_record, 'g_time')
        self.flush(self.vel_record, 'velocity')
        self.flush(self.pos_record, 'position')
        self.flush(self.teleport_record, 'teleport')
        self.flush(self.pause_record, 'paused')
        self.flush(self.input_record_1, 'input_1')
        self.flush(self.input_record_2, 'input_2')
        self.flush(self.output_record_1, 'output_1')
        self.flush(self.output_record_2, 'output_2')
        self.flush(self.output_record_3, 'output_3')
        self.flush(self.output_record_4, 'output_4')

        for zone_type in self.zone_types:
            self.flush(
                self.zone_type_records[zone_type], "zone_types/" + zone_type)

        self.flush(self.zone_id_records, 'zone')

    def flush(self, record, fieldname):
        '''
        Writes the record list into the given field and clears it.

        :param record: The record that should be written to file.
        :type record: list or np.ndarray

        :param fieldname: The name of the field in the HDF5 file this record should 
            be written into.
        :type fieldname: str
        '''
        if type(record) is np.ndarray:
            new_count = record.shape[0] - self.vrl[fieldname].shape[0]
            self.vrl[fieldname].resize(record.shape[0], axis=0)
            self.vrl[fieldname][-new_count:] = record[-new_count:]
        else:
            self.vrl[fieldname].resize(
                self.vrl[fieldname].shape[0] + len(record), axis=0)
            self.vrl[fieldname][-len(record):] = record
            del record[:]

    def close(self):
        ''' Record the time and close the log. '''
        end_time = time.time()
        self.vrl.attrs['end_time'] = end_time
        self.vrl.attrs['end_time_hr'] = time.strftime(
            "%Y.%m.%d - %H:%M:%S", time.localtime(end_time))
        self.vrl.close()


class Session(object):
    '''
    A play/simulation session for a LinMaze Level.
    
    :param level: The LinMaze Level that will be played in this Session.
    :type level: Level
    
    :param vel_ratio: The velocity read from the Gramophone is multiplied 
        with this. 1 by default.
    :type vel_ratio: float
    
    :param runtime_limit: How long should the simulation run in minutes. 
        Set to None to run infinately. None by default
    :type runtime_limit: float or None
    
    :param left_monitor: The number of the monitor to the right of the animal. Set to 
        None to disable this monitor. 1 by default.
    :type left_monitor: int or None
    
    :param right_monitor: The number of the monitor to the right of the animal. Set to 
        None to disable this monitor. None by default.
    :type right_monitor: int or None
    
    :param gramophone_serial: The serial of the Gramophone used for the simulation.
        Set to None to find a Gramophone automatically. None by default.
    :type gramophone_serial: int or None
    
    :param fullscreen: Should the simulation run in fullscreen mode? True by default.
    :type fullscreen: bool
    
    :param offset_arrow: Should the zone_offset of the Level be shown as a red arrow on screen? 
        False by default.
    :type offset_arrow: bool
    
    :param skip_save: Should the saving of a log be skipped for this Session? False by default.
    :type skip_save: bool
    '''

    def __init__(self, level, vel_ratio=1, runtime_limit=None,
                 left_monitor=1, right_monitor=None, gramophone_serial=None,
                 fullscreen=True, offset_arrow=False, skip_save=False):

        self.level = level
        self.vel_ratio = vel_ratio
        self.runtime_limit = runtime_limit
        self.left_monitor = left_monitor
        self.right_monitor = right_monitor
        self.gramophone_serial = gramophone_serial
        self.offset_arrow = offset_arrow
        self.skip_save = skip_save

        grams = Comms.find_devices()
        if grams:
            if self.gramophone_serial is None:
                print('\nNo Gramophone specified. Using the first one.')
                self.gramophone = grams[list(grams)[0]]
            else:
                self.gramophone = grams[self.gramophone_serial]
        else:
            raise(LinMazeError('No gramophones connected.'))

        self.vr_units = []
        self.runtime = Stopwatch()
        self.position = level.zone_offset - 1
        self.current_zone = self.level.zones[0]
        self.paused = False
        self.teleported = False
        self.last_position = 0

        # Render the level if it wasn't pre rendered
        if not level.rendered:
            level.render()

        # Set session for all events
        for key in self.level.events:
            self.level.events[key].set_session(self)

        # Make the window
        if left_monitor is not None:
            left_window = VRWindow(
                self, self.left_monitor, mirrored=False, fullscreen=fullscreen)
        if right_monitor is not None:
            right_window = VRWindow(
                self, self.right_monitor, mirrored=True, fullscreen=fullscreen)

        def main_loop(dt):
            '''
            Commands executed at each frame refresh.
            
            :param dt: Time since last reftesh is seconds. Passed by the pyglet clock.
            :type dt: float
            '''
            # print('FPS:', 1/dt)s
            self.gramophone.read_linmaze_params()
            velocity = round(
                self.vel_ratio*(self.gramophone.last_position - self.last_position)/14400)
            self.last_position = self.gramophone.last_position
            g_time, in_1, in_2, out_1, out_2, out_3, out_4 = \
                self.gramophone.last_time,  \
                self.gramophone.last_in_1,  \
                self.gramophone.last_in_2,  \
                self.gramophone.last_out_1, \
                self.gramophone.last_out_2, \
                self.gramophone.last_out_3, \
                self.gramophone.last_out_4


            if not self.paused:
                self.movement(velocity)

            self.check_zone()
            self.check_rules(velocity)

            if not self.skip_save:
                self.log.make_entry(velocity, g_time, in_1, in_2, out_1, out_2, out_3, out_4)

            if self.runtime_limit is not None and\
                    self.runtime.value() >= self.runtime_limit * 60:
                pyglet.app.exit()

        # pyglet.clock.set_fps_limit(30)
        pyglet.clock.schedule(main_loop)
        # pyglet.clock.schedule_interval(main_loop, 0.005)

        # Make an OpenGL texture from every frame's texture
        texture_ids = []
        textures = [frame.texture for frame in self.level.frames]
        textures.append(self.level.dummy_frame.texture)
        for tex in textures:
            # from PIL import Image
            # Image.fromarray(tex).show()
            sy = tex.shape[0]
            sx = tex.shape[1]
            # print('X', tex.shape[1], 'Y', tex.shape[0], 'C', tex.shape[2])

            tid = GLuint(0)
            glGenTextures(1, ctypes.byref(tid))

            glPixelStorei(GL_UNPACK_ALIGNMENT, 1)
            glBindTexture(GL_TEXTURE_2D, tid)
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, sx, sy,
                         0, GL_RGB, GL_UNSIGNED_BYTE, tex.ctypes.data)
            glGenerateMipmap(GL_TEXTURE_2D)
            texture_ids.append(tid)

        # Make VR units
        for i, frame in enumerate(self.level.frames):
            self.vr_units.append(
                {"length": frame.width - self.level.transition_width,
                 "texture_id": texture_ids[i],
                 "position": self.level.zones[i].begin})

        # VR unit of dummy frame
        self.vr_units.append(
            {"length": self.level.dummy_frame.width,
             "texture_id": texture_ids[-1],
             "position": self.level.zones[-1].end})

        # Calculate level length
        self.virtual_length = 0
        for vru in self.vr_units:
            self.virtual_length += vru["length"]

        # Connect Gramophone, and reset outputs to 0
        self.gramophone.open()
        self.gramophone.write_output(1, 0)
        self.gramophone.write_output(2, 0)
        self.gramophone.write_output(3, 0)
        self.gramophone.write_output(4, 0)
        self.gramophone.write_analog(0)

        # Save start date and time
        self.start_time = time.time()
        self.start_time_hr = time.strftime(
            "%Y.%m.%d - %H:%M:%S", time.localtime(self.start_time))
        default_filename = time.strftime(
            "%Y-%m-%d %H_%M", time.localtime(self.start_time)) +\
            ' (' + self.level.name + ')'

        # Set up VR logger
        self.filename = None
        if not self.skip_save:
            self.filename = select_file(defaultextension='.vrl',
                                        filetypes=[('VR Log', '.vrl')],
                                        title='Save log for this session',
                                        initialdir=os.getcwd(),
                                        initialfile=default_filename)
            self.folder = os.path.dirname(os.path.realpath(self.filename))
            self.log = VRLog(self)

        # Show the window
        if left_monitor is not None:
            left_window.set_visible()
        if right_monitor is not None:
            right_window.set_visible()

        # Reset runtime & display start message
        self.runtime.reset()
        print("Starting VR session \n  Level: " + self.level.name +
              "\n  Date & Time: " + self.start_time_hr +
              "\n  Log file: " + str(self.filename) + "\n")

        if self.runtime_limit is not None:
            finish_time = self.start_time + self.runtime_limit * 60
            print("Training is limited to", self.runtime_limit,
                  'minutes. It will automatically end at',
                  time.strftime("%H:%M", time.localtime(finish_time)), '\n')

        # Main app run
        pyglet.app.run()

        # After main app is closed
        # Reset outputs to 0 and disconnect the Gramophone
        self.gramophone.stop_burst(1)
        self.gramophone.stop_burst(2)
        self.gramophone.stop_burst(3)
        self.gramophone.stop_burst(4)
        self.gramophone.write_output(1, 0)
        self.gramophone.write_output(2, 0)
        self.gramophone.write_output(3, 0)
        self.gramophone.write_output(4, 0)
        self.gramophone.write_analog(0)
        self.gramophone.close()

        # Save all remaining data
        if not self.skip_save:
            self.log.flush_all()
            self.log.close()

    def movement(self, vel):
        '''
        Move on the map with given velocity.
        
        :param vel: Distance to move in pixels.
        :type vel: int
        '''

        # Base movement (used to calculate others, loops around)
        self.position += vel
        self.position %= -self.virtual_length + self.level.screen_width

        # Image movement
        img_move = self.position - self.vr_units[0]["position"]
        for vru in self.vr_units:
            vru["position"] += img_move

        # Virtual movement (position of the "character")
        limit = self.virtual_length - self.level.screen_width \
            - self.level.zone_offset

        if -self.position > limit:
            self.virtual_position = -self.position - \
                (self.virtual_length - self.level.screen_width -
                 self.level.zone_offset)
        else:
            self.virtual_position = -self.position + self.level.zone_offset

    def pause(self, position=None):
        '''
        Pauses the level at the given position.
        
        :param position: Where should the simulation pause on the Level in pixels.
            Set to None to pause at current position. None by default.
        :type position: int or None
        '''

        if not self.paused:
            if position is not None:
                self.teleport(position)
            self.paused = True

    def unpause(self, position=None):
        '''
        Unpauses the level at the given position.
        
        :param position: Where should the simulation unpause on the Level in pixels.
            Set to None to unpause at current position. None by default.
        :type position: int or None
        '''
        if self.paused:
            if position is not None:
                self.teleport(position)
            # for zr in self.zone_rules:
            #     zr.delay_timer.reset()
            self.paused = False

    def teleport(self, target_pos):
        '''
        Teleports to the given position.

        :param target_pos: Where should the teleportation land in pixels.
        :type position: int
        
        '''
        self.position = -(target_pos - self.level.zone_offset)
        self.teleported = True

    def random_teleport(self, target_zone_types):
        '''
        Teleports to the middle of a random zone with one of the given zone types.
        
        :param target_zone_types: list of possible landing zone types.
        :type target_zone_types: [str]
        '''
        zone_selection = [
            zone for zone in self.level.zones
            if zone.zone_type in target_zone_types]
        #and zone.zone_type != self.current_zone.zone_type

        target_zone = random.choice(zone_selection)
        middle_of_target = (target_zone.begin + target_zone.end) // 2

        self.teleport(middle_of_target)

    def check_zone(self):
        ''' Updates the current zone. '''
        self.current_zone = [zone for zone in self.level.zones
                             if zone.check(self.virtual_position)][0]

    def check_rules(self, vel):
        '''
        Checks all the rules of the Level.
        
        :param vel: The current velocity (for velocity based rules).
        :type vel: int
        '''

        for rule in self.level.rules:
            # Check zone rules
            if type(rule) is Rule.ZoneRule:
                rule.check(self.current_zone.zone_type)
            # Check speed and velocity rules
            if type(rule) in [Rule.SpeedRule, Rule.VelocityRule, Rule.SmoothVelocityRule]:
                rule.check(vel)



import struct
import threading
from collections import namedtuple
from random import randint, sample
from threading import Thread
from time import sleep, time

from PyQt5.QtCore import QObject, QThread, pyqtSignal, pyqtSlot
from pywinusb import hid


def background(fn):
    """ Decorator for functions that should run in the background. """
    def run(*k, **kw):
        t = threading.Thread(target=fn, args=k, kwargs=kw)
        t.start()
        return t
    return run

def find_devices():
    dev_filter = hid.HidDeviceFilter(vendor_id=0x0483, product_id=0x5750)
    devices = {}
    for dev in dev_filter.get_devices():
        gram = Gramophone(dev, verbose=False)
        gram.open()
        for _ in range(5):
            gram.read_sensors()
            gram.read_firmware_info()
            gram.read_product_info()
            sleep(0.1)
            if gram.product_name == 'GRAMO-01':
                devices[gram.product_serial] = gram
                break
        gram.close()
    return devices

class Transmitter(QObject):
    """
    Emits Qt signals to transmit infromation from incoming packets.
    """
    velocity_signal = pyqtSignal(float)
    position_signal = pyqtSignal(int)
    position_diff_signal = pyqtSignal(int)
    inputs_signal = pyqtSignal(int, int)
    recorder_signal = pyqtSignal(int, float, int, int)
    device_error = pyqtSignal(str)

    def __init__(self):
        super().__init__()
        self.last_pos = 0

    def emit_velocity(self, vel):
        """ 
        Emit a signal with the given velocity.

        :param vel: The velocity that should be emitted.
        :type vel: float
        """
        self.velocity_signal.emit(vel)

    def emit_position(self, pos):
        """ 
        Emit a signal with the given position.

        :param pos: The position that should be emitted.
        :type pos: int
        """
        self.position_signal.emit(pos)
        self.position_diff_signal.emit(pos-self.last_pos)
        self.last_pos = pos

    def emit_inputs(self, inputs):
        """ 
        Emit a signal with the state of the digital inputs.

        :param inputs: A list containing the states of the two digital inputs.
        :type inputs: [int, int]
        """
        self.inputs_signal.emit(inputs[0], inputs[1])

    def emit_recorder(self, values):
        """ 
        Emit a signal with all the data required by the Recorder module.
        These are the current time, velocity and input states in order.

        :param values: A list containing the time, velocity and input states.
        :type values: [int, float, int, int]
        """
        self.recorder_signal.emit(values[0], values[1],
                                  values[2], values[3])

    def emit_device_error(self, error_msg):
        """ 
        Emit a signal with the given error message.

        :param error_msg: The error message.
        :type error_msg: str
        """
        self.device_error.emit(str(error_msg))


class Reader(QObject):
    """
    A worker that can continnously send commands to the Gramophone with a given frequency.

    :param name: The name of this reader. Can be used for identification to stop a specific reader.
    :type name: str

    :param read_func: The read function that should be called.
    :type read_func: function

    :param frequency: The frequency at which the function should be called in Hz.
    :type frequency: float
    """
    def __init__(self, name, read_func, frequency):
        super().__init__()
        self.name = name
        self.read_func = read_func
        self.reading = None
        self.frequency = frequency

    @pyqtSlot()
    def read(self):
        """ Start calling the read function repeatedly. Slot for a QThread. """
        self.reading = True
        while self.reading:
            self.read_func()
            sleep(1/self.frequency)

    def abort(self):
        """ Stop calling the read function. """
        self.reading = False


class Gramophone(hid.HidDevice):

    error_codes = {0x00: 'PACKET_FAIL_UNKNOWNCMD',
                   0x01: 'PACKET_FAIL_INVALIDCMDSYNTAX',
                   0x04: 'PACKET_FAIL_INVALIDPARAMSYNTAX',
                   0x05: 'PACKET_FAIL_RANGEERROR',
                   0x06: 'PACKET_FAIL_PARAMNOTFOUND',
                   0x07: 'PACKET_FAIL_VALIDFAIL',
                   0x08: 'PACKET_FAIL_ACCESSVIOLATION'}

    Parameter = namedtuple('Parameter', ['name', 'info', 'type'])
    parameters = {0x01: Parameter('VSEN3V3', 'Voltage on 3.3V rail.', 'float'),
                  0x02: Parameter('VSEN5V', 'Voltage on 5V rail.', 'float'),
                  0x03: Parameter('TSENMCU', 'Internal teperature of the MCU.', 'float'),
                  0x04: Parameter('TSENEXT', 'External temperature sensor on the board.', 'float'),
                  0x0A: Parameter('SENSORS', 'The voltages and temperatures in one parameter', 'float_list'),
                  0x05: Parameter('TIME', 'The time of the internal clock of the device.', 'uint64'),
                  0x10: Parameter('ENCPOS', 'Encoder position.', '-int32'),
                  0x11: Parameter('ENCVEL', 'Encoder velocity.', 'vel_struct'),
                  0x12: Parameter('ENCVELWIN', 'Encoder velocity window size.', 'uint16'),
                  0x13: Parameter('ENCHOME', 'Encoder homing.', 'uint8'),
                  0x14: Parameter('ENCHOMEPOS', 'Encoder home position.', '-int32'),
                  0x20: Parameter('DI-1', 'Digital input 1.', 'uint8'),
                  0x21: Parameter('DI-2', 'Digital input 2.', 'uint8'),
                  0x25: Parameter('DI', 'Digital inputs.', 'list'),
                  0x30: Parameter('DO-1', 'Digital output 1.', 'uint8'),
                  0x31: Parameter('DO-2', 'Digital output 2.', 'uint8'),
                  0x32: Parameter('DO-3', 'Digital output 3.', 'uint8'),
                  0x33: Parameter('DO-4', 'Digital output 4.', 'uint8'),
                  0x35: Parameter('DO', 'Digital outputs.', 'list'),
                  0x40: Parameter('AO', 'Analogue output.', 'float'),
                  0xFF: Parameter('LED', 'LED state changed', 'uint8'),
                  0xAA: Parameter('REC', 'Bundle of parameters for the Recorder module.', 'recorder'),
                  0xBB: Parameter('LINM', 'Bundle of parameters for the LinMaze module.', 'linmaze')
                  }


    def __init__(self, device, verbose=False):
        self.device = device
        super().__init__(self.device.device_path,
                         self.device.parent_instance_id,
                         self.device.instance_id)

        self.verbose = verbose

        self.report = None
        self.set_raw_data_handler(self.data_handler)

        self.transmitter = Transmitter()
        self.readers = []

        self.target = [randint(0x00, 0xFF), randint(0x00, 0xFF)]
        self.source = [randint(0x00, 0xFF), randint(0x00, 0xFF)]

        self.ping_time = None
        self.app_state = 'Unknown'
        self.is_open = False

        self.firmware_release = None
        self.firmware_sub = None
        self.firmware_build = None
        self.firmware_year = None
        self.firmware_month = None
        self.firmware_day = None
        self.firmware_hour = None
        self.firmware_minute = None
        self.firmware_second = None

        self.product_name = None
        self.product_revision = None
        self.product_serial = None
        self.product_year = None
        self.product_month = None
        self.product_day = None

        self.last_position = 0
        self.last_time = 0
        self.last_velocity = 0
        self.last_in_1 = 0
        self.last_in_2 = 0
        self.last_out_1 = 0
        self.last_out_2 = 0
        self.last_out_3 = 0
        self.last_out_4 = 0

        self.bursting = {1: False,
                         2: False,
                         3: False,
                         4: False}
        self.burst_threads = []

        self.sensor_values = {'VSEN3V3': None,
                              'VSEN5V': None,
                              'TSENMCU': None,
                              'TSENEXT': None}

    def open(self):
        """ Connect to this device. """
        super().open()
        reports = self.find_output_reports()
        self.report = reports[0]
        self.check_app()
        sleep(0.1)
        if self.app_state == 'IAP':
            print('Device in IAP state. Resetting...')
            self.reset()
            sleep(0.1)
        self.is_open = True

    def close(self):
        """ Disconnect from this device. """
        self.report = None
        super().close()
        self.is_open = False

    @staticmethod
    def decode_payload(val_type, payload):
        if val_type == 'float':
            return struct.unpack('f', payload)[0]
        if val_type == 'float_list':
            float_list = []
            for I in range(len(payload)//4):
                float_list.append(struct.unpack('f', payload[I*4:I*4+4])[0])
            return float_list
        if val_type == 'int32':
            return int.from_bytes(payload, 'little', signed=True)
        if val_type == '-int32':
            return -int.from_bytes(payload, 'little', signed=True)
        if val_type in ['uint8', 'uint16', 'uint64']:
            return int.from_bytes(payload, 'little', signed=False)
        if val_type == 'vel_struct':
            return -struct.unpack('f', payload[0:4])[0]*float(payload[4])
        if val_type == 'list':
            return list(payload)
        if val_type == 'recorder':
            recorder_data = [] # time, vel, in_1, in_2
            recorder_data.append(int.from_bytes(
                payload[0:8], 'little', signed=False))
            recorder_data.append(-struct.unpack(
                'f', payload[8:12])[0]*float(payload[12]))
            recorder_data.append(payload[13])
            recorder_data.append(payload[14])
            return recorder_data
        if val_type == 'linmaze':
            linmaze_data = [] # time, pos, in_1, in_2, out_1, out_2, out_3, out_4
            linmaze_data.append(int.from_bytes(
                payload[0:8], 'little', signed=False))
            linmaze_data.append(-int.from_bytes(
                payload[8:12], 'little', signed=True))
            linmaze_data.append(payload[12])
            linmaze_data.append(payload[13])

            linmaze_data.append(payload[14])
            linmaze_data.append(payload[15])
            linmaze_data.append(payload[16])
            linmaze_data.append(payload[17])
            return linmaze_data

        return None

    def read_input(self, input_id):
        self.read_param(0x20+input_id-1)

    def read_inputs(self):
        self.read_params(0x25, [0x20, 0x21])

    def read_output(self, output_id):
        self.read_param(0x30+output_id-1)

    def read_outputs(self):
        self.read_params(0x35, [0x30, 0x31, 0x32, 0x33])

    def read_analog_out(self):
        self.read_param(0x40)

    def read_voltages(self):
        self.read_param(0x01)
        self.read_param(0x02)

    def read_temperatures(self):
        self.read_param(0x03)
        self.read_param(0x04)

    def read_sensors(self):
        self.read_params(0x0A, [0x01, 0x02, 0x03, 0x04])

    def read_time(self):
        self.read_param(0x05)

    def read_position(self):
        self.read_param(0x10)

    def read_velocity(self):
        self.read_param(0x11)

    def read_window_size(self):
        self.read_param(0x12)

    def read_homing_state(self):
        self.read_param(0x13)
        self.read_param(0x14)

    def read_firmware_info(self):
        self.send(0x00, 0x04, [])

    def read_product_info(self):
        self.send(0x00, 0x08, [])

    def read_recorder_params(self):
        self.read_params(0xAA, [0x05, 0x11, 0x20, 0x21])

    def read_linmaze_params(self):
        self.read_params(0xBB, [0x05, 0x10, 0x20, 0x21, 0x30, 0x31, 0x32, 0x33])

    def write_output(self, output, value):
        self.write_param(0x30+output-1, [int(value)])

    def write_analog(self, value):
        self.write_param(0x40, list(struct.pack('f', value)))

    def ping(self):
        data = sample(range(0, 255), 5)
        self.ping_time = time()
        self.send(0x00, 0x00, data)
        print('Ping!', data)

    def check_app(self):
        self.send(0x00, 0x05, [])

    def set_LED(self, state):
        self.send(0xFF, 0x12, [state])

    def reset(self):
        self.send(0x00, 0xF0, [])

    def reset_time(self):
        self.write_param(0x05, [0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00])

    def reset_position(self):
        self.write_param(0x10, [0x00, 0x00, 0x00, 0x00])

    def read_param(self, param):
        self.send(param, 0x0B, [param])

    def read_params(self, msn, params):
        self.send(msn, 0x0B, params)

    def write_param(self, param, payload):
        self.send(param, 0x0C, [param]+payload)

    def send(self, msn, cmd, payload):
        plen = len(payload)
        filler = [0] * (65-plen-8)
        full = [0x0] + self.target + self.source + \
            [msn, cmd, plen] + payload + filler
        try:
            self.report.set_raw_data(full)
            self.report.send()
        except hid.helpers.HIDError as hid_error:
            # print('HID ERROR', hid_error)
            self.transmitter.emit_device_error(hid_error)

        # print('Sent:',[hex(byte) for byte in full])

        # print('Sent')
        # print('Target', self.target)
        # print('Source', self.source)
        # print('MSN', msn)
        # print('CMD', cmd)
        # print('PLen', plen)
        # print('Payload', payload)
        # print()

    def data_handler(self, data):
        target = data[1:3]
        source = data[3:5]
        if target == self.source and source == self.target:
            msn = data[5]
            cmd = data[6]
            plen = data[7]
            payload = data[8:8+plen]

            if cmd == 0x00:
                delay = time()-self.ping_time
                print('Pong!', payload)
                print('Delay:', delay, 'sec\n')

            if cmd == 0x01:
                if self.verbose:
                    print(Gramophone.parameters[msn].name, 'OK!')

            if cmd == 0x02:
                print(Gramophone.parameters[msn].name, 'failed.',
                      self.error_codes[payload[0]])

            if cmd == 0x04:
                self.firmware_release = payload[0]
                self.firmware_sub = payload[1]
                self.firmware_build = int.from_bytes(
                    payload[2:4], 'little', signed=False)
                self.firmware_year = int.from_bytes(
                    payload[4:6], 'little', signed=False)
                self.firmware_month = payload[6]
                self.firmware_day = payload[7]
                self.firmware_hour = payload[8]
                self.firmware_minute = payload[9]
                self.firmware_second = payload[10]

                if self.verbose:
                    print('Firmware version')
                    print('Relase:', str(self.firmware_release) +
                          '.'+str(self.firmware_sub))
                    print('Build:', self.firmware_build)
                    print('Date:', str(self.firmware_year)+'-' +
                          str(self.firmware_month)+'-'+str(self.firmware_day))
                    print('Time', str(self.firmware_hour)+':' +
                          str(self.firmware_minute)+':'+str(self.firmware_second))
                    print()

            if cmd == 0x05:
                if payload[0] == 0x00:
                    if self.verbose:
                        print('CheckApp: IAP \n')
                    self.app_state = 'IAP'
                if payload[0] == 0x01:
                    if self.verbose:
                        print('CheckApp: Application \n')
                    self.app_state = 'App'

            if cmd == 0x08:
                self.product_name = ''.join(
                    [chr(byte) for byte in payload[0:18] if byte != 0x00])
                self.product_revision = ''.join(
                    [chr(byte) for byte in payload[18:24]])
                self.product_serial = int.from_bytes(
                    payload[24:28], 'little', signed=False)
                self.product_year = int.from_bytes(
                    payload[28:30], 'little', signed=False)
                self.product_month = int.from_bytes(
                    payload[30:31], 'little', signed=False)
                self.product_day = int.from_bytes(
                    payload[31:32], 'little', signed=False)

                if self.verbose:
                    print('Product Info')
                    print('Name:', self.product_name)
                    print('Revision:', self.product_revision)
                    print('Serial', hex(self.product_serial))
                    print('Production:', str(self.product_year)+'-' +
                          str(self.product_month)+'-'+str(self.product_day))
                    print()

            if cmd == 0x0B:
                val = None
                if payload is not None:
                    val = Gramophone.decode_payload(
                        Gramophone.parameters[msn].type, bytes(payload))
                if self.verbose:
                    print('Read:', Gramophone.parameters[msn].info,
                          'Value:', val)

                if Gramophone.parameters[msn].name == 'ENCVEL':
                    self.transmitter.emit_velocity(val)
                    self.last_velocity = val
                if Gramophone.parameters[msn].name == 'ENCPOS':
                    self.transmitter.emit_position(val)
                    self.last_position = val
                if Gramophone.parameters[msn].name == 'SENSORS':
                    self.sensor_values['VSEN3V3'] = val[0]
                    self.sensor_values['VSEN5V'] = val[1]
                    self.sensor_values['TSENMCU'] = val[2]
                    self.sensor_values['TSENEXT'] = val[3]
                if Gramophone.parameters[msn].name == 'DI':
                    self.transmitter.emit_inputs(val)
                    self.last_in_1 = val[0]
                    self.last_in_2 = val[1]
                if Gramophone.parameters[msn].name == 'DO':
                    self.last_out_1 = val[0]
                    self.last_out_2 = val[1]
                    self.last_out_3 = val[2]
                    self.last_out_4 = val[3]
                if Gramophone.parameters[msn].name == 'REC':
                    self.transmitter.emit_recorder(val)
                    self.last_time = val[0]
                    self.last_velocity = val[1]
                    self.last_in_1 = val[2]
                    self.last_in_2 = val[3]
                if Gramophone.parameters[msn].name == 'LINM':
                    self.transmitter.emit_recorder(val)
                    self.last_time = val[0]
                    self.last_position = val[1]
                    self.last_in_1 = val[2]
                    self.last_in_2 = val[3]
                    self.last_out_1 = val[4]
                    self.last_out_2 = val[5]
                    self.last_out_3 = val[6]
                    self.last_out_4 = val[7]

            if cmd not in [0x00, 0x01, 0x02, 0x04, 0x05, 0x08, 0x0B]:
                print('CMD', hex(cmd))
                print('plen', plen)
                print('payload', payload)

    @background
    def start_burst(self, port, on_time, pause_time):
        self.bursting[port] = True
        while self.bursting[port]:
            self.write_output(port, 1)
            sleep(on_time)
            self.write_output(port, 0)
            sleep(pause_time)

    def stop_burst(self, port):
        self.bursting[port] = False

    def start_reader(self, name, param, freq):
        command = {'position': self.read_position,
                   'velocity': self.read_velocity,
                   'time': self.read_time,
                   'inputs': self.read_inputs,
                   'recorder': self.read_recorder_params
                   }[param]
        reader = Reader(name, command, freq)
        thread = QThread()
        # thread.setObjectName('thread_' + str(idx))
        self.readers.append((thread, reader))
        reader.moveToThread(thread)

        thread.started.connect(reader.read)
        thread.start()

    def stop_reader(self, name=None):
        if self.readers:
            for thread, reader in self.readers:
                if name is None or reader.name == name:
                    reader.abort()
                    thread.quit()
                    thread.wait()



import time

class Timer(object):
	def __init__(self, length=0):
		self.length = length
		self.until = time.time()+self.length
	
	def __str__(self):
		return "%5.2f sec" %(self.until-time.time())
		
	def is_running(self):
		return time.time() < self.until
	
	def set(self, length):
		self.length = length
		
	def reset(self):
		self.until = time.time()+self.length


''' Tools for reading and writing pyVR related files '''

import csv
import os


def save_lists(filename, headers, lists, delimiter=","):
    ''' Writes the provided lists into a csv '''

    with open(filename, "w", newline='') as csv_file:
        csv_file.write('sep=' + delimiter + os.linesep)
        writer = csv.writer(csv_file)
        writer.writerow(headers)
        writer.writerows(zip(*lists))
        # for row in list(zip(*lists)):
        #     print(row)

    print('Data written to: ' + filename + '\n')


def select_file(*args, **kwargs):
    ''' Displays a file selection dialog and returns the file selected '''
    ''' eg. defaultextension='.csv', filetypes=[('VR Log (.csv)', '.csv')],
        title='Save log for this session', initialdir='C:/VR/Results',
        initialfile=self.start_datetime_str '''

    import tkinter
    from tkinter import filedialog
    root = tkinter.Tk()
    root.withdraw()
    file = filedialog.asksaveasfilename(parent=root, *args, **kwargs)

    return str(file)



''' Contains the Zone object '''


class Zone(object):
    '''
    Correlates with beginning and end of Frames to add functionality.
    
    :param begin: The beginning of the Zone in pixels.
    :type begin: int

    :param end: The end of the Zone in pixels.
    :type end: int

    :param zone_type: The type of the Zone for Rules. 'generic' by default.
    :type zone_type: str
    '''
    zone_id = 0

    def __init__(self, begin, end, zone_type='generic'):
        Zone.zone_id += 1

        self.zone_id = Zone.zone_id
        self.begin = begin
        self.end = end
        self.zone_type = zone_type

    def __str__(self):
        return "Zone  #" + str(self.zone_id).ljust(2) + " - Range: [" + str(self.begin) + "," + str(self.end) + "] - Type: " + str(self.zone_type)

    def check(self, pos):
        '''
        Returns True if the given position is in the zone.
        
        :param pos: The position in pixels.
        :type pos: int

        :rtype: bool
        '''
        return self.begin <= pos <= self.end



from . import Comms
from . import Recorder
from . import LinMaze
import os
import sys
import subprocess

__version__ = '0.5.18'

def main(args=None):
    """ The main routine. """
    project_dir = os.path.split(__file__)[0]

    if args is None:
        args = sys.argv[1:]
    
    if args:
        if args[0] == 'examples':
            # subprocess.Popen(r'explorer "{}\examples"'.format(project_dir))
            os.startfile(project_dir+r'\examples')
        if args[0] == 'guide':
            os.startfile(project_dir+r'\docs\_downloads\Gramophone User Guide.pdf')
        if args[0] == 'docs':
            os.startfile(project_dir+r'\docs\index.html')
        if args[0] == 'ver':
            print('\n Gramophone Tools version:', __version__)
            input(' Press ENTER to continnue...')
        if args[0] not in ['examples', 'guide', 'docs', 'ver']:
            help_text()
    else:
        help_text()

def help_text():
    print('\n Use one of the following arguments:\n')
    print('    ver - Displays the current version of the GramophoneTools package in the command line.')
    print('    guide - Opens the user guide of the Gramophone system.')
    print('    docs - Opens the documentation of the GramophoneTools package.')
    print('    examples - Opens a folder containing examples for the LinMaze submodule.')
    print('\n eg.: gram guide\n')
    input(' Press ENTER to continnue...')


''' Triggers an event on a given criteria '''
from statistics import mean
from collections import deque
from abc import ABC, abstractmethod

import numpy as np

from GramophoneTools.LinMaze import Event
from GramophoneTools.LinMaze.Tools.Timer import Timer

class Rule(ABC):
    '''
    Generic Rule all specific Rules intherit from.
        
    :param level: The Level this Rule is active on.
    :type level: Level

    :param event: The event the rule triggers.
    :type event: Event
    '''

    def __init__(self, level, event):
        self.level = level
        self.event = event
        self.done = False

    def trigger(self):
        ''' Triggers the associated event and prints out a message about it. '''
        if self.event.triggerable:
            print(self, "rule triggered")
            self.event.trigger()
            self.done = True
            # print(self.event)
    
    @abstractmethod
    def check(self):
        ''' Check whether the rules's event should be triggered. '''
        pass


class ZoneRule(Rule):
    ''' 
    A Rule that triggers if the animal is in a given type of zone.
        
    :param level: The Level this Rule is active on.
    :type level: Level

    :param event: The event the rule triggers.
    :type event: Event
    
    :param zone_type: The type of Zone this Rule is active in.
    :type zone_type: str
    
    :param delay: How many seconds should be spent in the Zone before triggering.
    :type delay: float
    '''

    def __init__(self, level, event, zone_type, delay):
        super().__init__(level, event)
        self.zone_type = zone_type
        self.delay = delay
        self.delay_timer = Timer(delay)
        self.active = False

    def __str__(self):
        return "In " + str(self.zone_type) + " zone for " + str(self.delay) + " sec"

    def check(self, current_zone_type):
        """ 
        Check whether the Zone rule should be triggered.
        
        :param current_zone_type: Type of the curren zone.
        :type current_zone_type: str
        """
        # print('Checking ',self)
        # print('CZ',current_zone_type)
        if current_zone_type == self.zone_type:
            if not self.active and not self.done:
                # Zone entry >> Rule activation
                self.delay_timer.reset()
                self.active = True
            if self.active and self.delay_timer.is_running() and not self.done:
                # Waiting for timer to run out
                pass
            if self.active and not self.delay_timer.is_running() and not self.done:
                # Timer ran out, trigger event
                self.trigger()
                if type(self.event) == Event.RandomTeleport:
                    self.done = False
                    self.active = False
        else:
            # print('current',current_zone_type,'not in',self.zone_type)
            self.done = False
            self.active = False


class VelocityRule(Rule):
    '''
    A Rule that triggers if the velocity is above or below a certain threshold.
    
    :param level: The Level this Rule is active on.
    :type level: Level

    :param event: The event the rule triggers.
    :type event: Event

    :param vel_rule_type: Type of comparison. Can be 'above' or 'below'.
    :type vel_rule_type: str

    :param threshold: Absolute velocity should be above or below this value.
    :type threshold: float

    :param delay: How long should the absolute velocity be above or below the threshold.
    :type delay: float    
    '''

    def __init__(self, level, event, vel_rule_type, threshold, delay):
        super().__init__(level, event)
        self.vel_rule_type = vel_rule_type
        self.threshold = threshold
        self.delay = delay
        self.delay_timer = Timer(delay)
        self.active = False

    def __str__(self):
        return "Velocity " + str(self.vel_rule_type) + " " + str(self.threshold) +\
            " for " + str(self.delay) + " sec"

    def check(self, vel):
        """ 
        Check whether the Velocity rule should be triggered.
        
        :param vel: The current velocity.
        :type vel: int
        """
        if self.vel_rule_type == "above":
            if abs(vel) > self.threshold:
                self.active = True
            if abs(vel) < self.threshold:
                self.active = False
                self.done = False
                self.delay_timer.reset()

        if self.vel_rule_type == "below":
            if abs(vel) < self.threshold:
                self.active = True
            if abs(vel) > self.threshold:
                self.active = False
                self.done = False
                self.delay_timer.reset()

        if self.active and not self.done and not self.delay_timer.is_running():
            self.trigger()
            # self.delay_timer.reset()

class SmoothVelocityRule(VelocityRule):
    '''
    A Rule that triggers if the moveing average of velocity is above or below a certain threshold.
        
    :param level: The Level this Rule is active on.
    :type level: Level

    :param event: The event the rule triggers.
    :type event: Event

    :param bin_size: How many velocities should be used for calculating the moving average.
    :type bin_size: int

    :param vel_rule_type: Type of comparison. Can be 'above' or 'below'.
    :type vel_rule_type: str

    :param threshold: Smoothed absolute velocity should be above or below this value.
    :type threshold: float

    :param delay: How long should the smoothed absolute velocity be above or below the threshold.
    :type delay: float
    '''

    def __init__(self, level, event, bin_size, vel_rule_type, threshold, delay):
        super().__init__(level, event, vel_rule_type, threshold, delay)
        self.bin_size = bin_size
        self.vels = deque([], bin_size)

    def __str__(self):
        return "Smooth velocity (avg. of "+str(self.bin_size)+") " \
            + str(self.vel_rule_type) + " " + str(self.threshold) \
            + " for " + str(self.delay) + " sec"

    def check(self, vel):
        """ 
        Check whether the Smooth velocity rule should be triggered.
        
        :param vel: The current velocity.
        :type vel: int
        """
        self.vels.append(vel)
        smooth_vel = mean(self.vels)
        super().check(smooth_vel)


class SpeedRule(Rule):
    '''
    A Rule that triggers if the absolute integral of the velocity on a given range is above or below a given threshold.
        
    :param level: The Level this Rule is active on.
    :type level: Level

    :param event: The event the rule triggers.
    :type event: Event

    :param speed_rule_type: Type of comparison. Can be 'above' or 'below'.
    :type speed_rule_type: str

    :param threshold: The calculated integral should be above or below this value.
    :type threshold: float
     
    :param bin_size: How many velocities should be used for the integral.
    :type bin_size: int
     '''

    def __init__(self, level, event, speed_rule_type, threshold, bin_size):
        super().__init__(level, event)
        self.speed_rule_type = speed_rule_type
        self.threshold = threshold
        self.bin_size = bin_size
        self.record = np.zeros(bin_size)

    def __str__(self):
        return "Absolute sum of the last " + str(self.bin_size) + " velocities " +\
            str(self.speed_rule_type) + " " + str(self.threshold)

    def check(self, vel):
        """ 
        Check whether the Speed rule should be triggered.
        
        :param vel: The current velocity.
        :type vel: int
        """
        self.record[:-1] = self.record[1:]
        self.record[-1] = abs(vel)
        norm = sum(self.record)

        if self.speed_rule_type == "above":
            if norm > self.threshold and not self.done:
                self.trigger()
            if norm <= self.threshold and self.done:
                self.done = False

        if self.speed_rule_type == "below":
            if norm <= self.threshold and not self.done:
                self.trigger()
            if norm > self.threshold and self.done:
                self.done = False

        # print('sum: ', norm)



import sys
import Recorder
import os.path

def main(args=None):
    """The main routine."""
    if args is None:
        args = sys.argv[1:]

    if 'devmode' in args:
        devmode = True
        print('Launching in developer mode...')
    else:
        devmode = False

    file_args = [f for f in args if os.path.isfile(f)]

    if file_args:
        log_file = file_args[0]
        print('Open file:', log_file)
    else:
        log_file = None

    Recorder.main(devmode=devmode, log_file=log_file)

if __name__ == "__main__":
    main()


# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import GramophoneTools

setup(
    name='GramophoneTools',
    version=GramophoneTools.__version__,
    description='Tools for Gramophone systems by Femtonics Ltd.',
    url='http://femtonics.eu/',
    author='Femtonics Ltd.',
    author_email='info@femtonics.eu',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Scientific/Engineering',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Programming Language :: Python :: 3.6'],
    license='GNU GPLv3',
    keywords='femtonics gramophone',
    project_urls={
        'GitHub': 'https://github.com/Femtonics/GramophoneTools',
        'Documentation': 'http://gramophonetools.readthedocs.io',
        'User Guide': 'http://gramophonetools.readthedocs.io/en/latest/_downloads/Gramophone%20User%20Guide.pdf'},
    packages=['GramophoneTools', 'GramophoneTools.Comms', 'GramophoneTools.LinMaze',
              'GramophoneTools.LinMaze.Tools', 'GramophoneTools.Recorder', 'GramophoneTools.examples', 'GramophoneTools.docs'],
    # py_modules=["GramophoneTools.ver"],
    package_dir={'GramophoneTools.examples': 'examples',
                 'GramophoneTools.docs': 'docs/build/html'},
    install_requires=[
        'h5py',
        'PyQt5',
        'numpy',
        'dill',
        'pillow',
        'pyglet',
        'pyqtgraph',
        'matplotlib',
        'opencv-python',
        'pywinusb',
        'xlsxwriter'],
    extras_require={
        'tests': [
            'pytest'],
        'docs': [
            'sphinx',
            'sphinx_rtd_theme']},
    python_requires='>=3',
    package_data={
        'GramophoneTools.LinMaze': ['res/icon.png', '*.dll'],
        'GramophoneTools.Recorder': ['*.ui'],
        'GramophoneTools.docs': ['*', '*/*', '*/*/*', '*/*/*/*']
    },
    include_package_data=True,
    entry_points={
        'gui_scripts': ['gramrec = GramophoneTools.Recorder.__main__:main'],
        'console_scripts': ['gramver = GramophoneTools:print_version', 'gram = GramophoneTools:main']
    }
)



''' Contains the Level object '''

import os
import numpy as np
from PIL import Image

from GramophoneTools.LinMaze import LinMaze, Frame, Event, Rule
from GramophoneTools.LinMaze.Zone import Zone
from GramophoneTools.LinMaze.Tools import Stopwatch, progressbar
from GramophoneTools.LinMaze.Tools.filehandler import select_file


class Level(object):
    ''' 
    An object that can be played for conditioning.
    You can add frames, events and rules to it to tune your conditioning. 
    
    :param name: The name this Level will be refered to by (eg. in the default filename when saving)
    :type name:  str

    :param screen_res: The resolution of the simulation window. Set it to the resolution of the monitor 
        if you are going to make the simulation full screen. eg.: (800,600)
    :type screen_res: int, int

    :param zone_offset: Zone offset in pixels, ie. what is the position of the mouse on the screen from the left.
    :type zone_offset: int

    :param transition_width: How long should the transition (gradual fade) shold be between zones in pixels.
    :type transition_width: int

    :param rgb: Red, Green and Blue levels of the frames as floats between 0 and 1.
    :type rgb: float, float, float
    '''

    def __init__(self, name, screen_res, zone_offset,
                 transition_width=100, rgb=(1, 1, 1)):
        print("\nCreating new VR level: ", name, "\n")
        self.name = name
        # self.screen_res = screen_res
        self.screen_width = screen_res[0]
        self.screen_height = screen_res[1]
        self.zone_offset = zone_offset
        self.transition_width = transition_width
        self.rgb = rgb

        self.rendered = False

        self.frames = []
        self.events = {}
        self.zones = []
        self.rules = []

    def __str__(self):
        level_string = self.name + '\nResolution: ' + str(self.screen_width)+'×'+str(self.screen_height) + \
            '\nZone offset: ' + str(self.zone_offset) +\
            '\nTransition width: ' + str(self.transition_width)+'\n'

        level_string += "\nBLOCKS\n"
        for frame, zone in zip(self.frames, self.zones):
            level_string += str(frame) + '\n' + str(zone) + '\n\n'

        level_string += "\nRULES\n"
        for rule in self.rules:
            level_string += 'If '+str(rule).lower() + \
                ' then '+str(rule.event).lower()+'\n'

        return level_string

    @property
    def dummy_frame(self):
        ''' A frame that looks like the begining of the level. '''
        dummy_frame = Frame.Frame(self.screen_width, self.screen_height)
        dummy_frame.texture = self.combined_frame.texture[:, :self.screen_width].astype(
            np.uint8)
        dummy_frame.made = True
        return dummy_frame

    @property
    def combined_frame(self):
        ''' All the frames combined left to right. '''
        if not self.rendered:
            self.render()

        return Frame.combine(self.frames)

    @property
    def image(self):
        ''' An image of the Level's combined frames. '''
        return self.combined_frame.make_img()

    def add_block(self, frame_type, *args, **kwargs):
        ''' Adds a block of the specified type and parameters to the level. '''

        frame_type = frame_type.lower()

        width = kwargs.get('length', 0) + self.transition_width
        height = self.screen_height
        wavelength = kwargs.get('wavelength', None)
        angle = kwargs.get('angle', None)
        random_seed = kwargs.get('random_seed', None)
        side_length = kwargs.get('side_length', None)
        filename = kwargs.get('filename', None)

        frame_maker = {  # Frame type switch
            'sine': lambda: Frame.SineWave(width, height, wavelength, angle),
            'square': lambda: Frame.SquareWave(width, height, wavelength, angle),
            'greynoise': lambda: Frame.GreyNoise(width, height, random_seed),
            'binarynoise': lambda: Frame.BinaryNoise(width, height, random_seed),
            'cloud': lambda: Frame.Cloud(width, height, random_seed),
            'marble': lambda: Frame.Marble(width, height, random_seed),
            'wood': lambda: Frame.Wood(width, height, random_seed),
            'checkerboard': lambda: Frame.Checkerboard(width, height, side_length),
            'image': lambda: Frame.ImageFile(height, filename)
        }.get(frame_type, None)

        if frame_maker is None:
            raise ValueError(
                'Error: \'{}\' is not a valid frame_type'.format(frame_type))
        frame = frame_maker()

        # New zone starts at the end of the last one
        if self.zones:
            begin = self.zones[-1].end
            end = self.zones[-1].end + frame.width - self.transition_width
        else:  # Fist zone starts at 0
            begin = 0
            end = frame.width - self.transition_width
        zone = Zone(begin, end, kwargs.get('zone_type', 'generic'))

        self.frames.append(frame)
        self.zones.append(zone)

        print("\nAdded block to " + self.name + ":")
        print(str(frame))
        print(str(zone))

    def add_event(self, name, event_type, *args):
        ''' 
        Adds an Event to the level which can be triggered by Rules.

        :param name: The name this Event can be referenced by
        :type name: string

        :param event_type: type of the event can be 'teleport', 'random_teleport', 
            'start_burst', 'stop_burst', 'port_on', 'port_off', 'pause', 'unpause'
        :type event_type: string

        :param args: The arguments for the Event of the given type.
        '''

        if event_type == "teleport":
            self.events[name] = Event.Teleport(self, args[0])

        if event_type == "random_teleport":
            self.events[name] = Event.RandomTeleport(self, args[0])

        if event_type == "start_burst":
            self.events[name] = Event.StartBurst(
                self, args[0], args[1], args[2])

        if event_type == "stop_burst":
            self.events[name] = Event.StopBurst(self, args[0])

        if event_type == "port_on":
            self.events[name] = Event.PortOn(self, args[0])

        if event_type == "port_off":
            self.events[name] = Event.PortOff(self, args[0])

        if event_type == "pause":
            self.events[name] = Event.Pause(self, args[0])

        if event_type == "unpause":
            self.events[name] = Event.UnPause(self, args[0])

    def add_rule(self, rule_type, event_name, *args):
        '''
        Adds a rule that triggers events based on animal behaviour.

        :param rule_type: The type of this rule, can be 'zone', 'velocity' or 'speed'.
        :type rule_type: string

        :param event_name: The name of the Event this Rule can trigger.
        :type event_name: string

        :param args: The arguments for the Rule of the given type.
        '''
        if rule_type == "zone":
            self.rules.append(Rule.ZoneRule(
                self, self.events[event_name], args[0], args[1]))
            # zone_type, delay

        if rule_type == "velocity":
            self.rules.append(Rule.VelocityRule(
                self, self.events[event_name], args[0], args[1], args[2]))
            # vel_rule_type, threshold, delay

        if rule_type == "smooth_velocity":
            self.rules.append(Rule.SmoothVelocityRule(
                self, self.events[event_name], args[0], args[1], args[2], args[3]))
            # bin_size, vel_rule_type, threshold, delay

        if rule_type == "speed":
            self.rules.append(Rule.SpeedRule(
                self, self.events[event_name], args[0], args[1], args[2]))
            # speed_rule_type, threshold, bin_size

    def render(self):
        ''' Renders all the frames of the level, making it ready to be played. '''
        if not self.rendered:
            print('\n'+self.name,  '>>\n')
            self.frames = Frame.multi_make(self.frames)
            self.frames = Frame.frame_transitions(self.frames, self.transition_width)

            loadtime = Stopwatch.Stopwatch()
            for count, frame in enumerate(self.frames):
                frame.make_texture()
                progressbar.printProgressBar(count, len(self.frames),
                                             prefix=' Loading  ',
                                             suffix='  (' + str(loadtime) + ')')
            progressbar.printProgressBar(len(self.frames), len(self.frames),
                                         prefix=' Loading  ',
                                         suffix='  (' + str(loadtime) + ')')
            print('\n')

            self.rendered = True

    def show_image(self):
        ''' Displays the level as a picture '''
        self.image.show()

    def save_image(self):
        ''' Saves the picture of the Level '''
        filename = select_file(defaultextension='.bmp',
                               filetypes=[('VR Level image', '.bmp')],
                               title='Save image of Level',
                               initialdir=os.getcwd(),
                               initialfile=self.name)
        self.image.convert('RGB').save(filename)

    def save_summary(self):
        ''' Saves a human readable summary of the Level '''
        filename = select_file(defaultextension='.txt',
                               filetypes=[('Text summary', '.txt')],
                               title='Save summary of Level',
                               initialdir=os.getcwd(),
                               initialfile=self.name)

        with open(filename, 'w') as summary_file:
            summary_file.write(str(self))

    def play(self, *args, **kwargs):
        '''
        Plays the Level by making a Session for it.

        :param args: Arguments of the Session created.
        :param kwargs: Keyword arguments of the Session created.

        '''
        try:
            LinMaze.Session(self, *args, **kwargs)
        except LinMaze.LinMazeError as err:
            print('\nERROR:', err)
            input()
